@@ -20,6 +20,6 @@ class DestinationExistsError(Error, RuntimeError):
 
 class JobsCorruptedError(Error, RuntimeError):
     "The state point manifest file of one or more jobs cannot be openend or is corrupted."
-    def __init_(self, job_ids):
+    def __init__(self, job_ids):
         self.job_ids = job_ids
         "The job id(s) of the corrupted job(s)."
